import os, stat
import subprocess
from odoo_env.__init__ import __version__
from odoo_env.messages import Msg
from odoo_env.odoo_conf import OdooConf

msg = Msg()


class Command:
    def __init__(self, parent, command=False, usr_msg=False, args=False,
                 client_name=False):
        """
        :param parent: El objeto OdooEnv que lo contiene por los parametros
        :param command: El comando a ejecutar en el shell
        :param usr_msg: El mensaje a mostrarle al usuario
        :param args: Argumentos para chequear, define si se ejecuta o no
        :return: El objeto Comando que se ejecutara luego
        """
        self._parent = parent
        self._command = command
        self._usr_msg = usr_msg
        self._args = args
        self._client_name = client_name

    def check(self):
        # si no tiene argumentos para chequear no requiere chequeo,
        # lo dejamos pasar
        if not self._args:
            return True

        # le pasamos el chequeo al objeto especifico
        return self.check_args()

    def check_args(self):
        raise NotImplementedError

    def execute(self):
        cmd = self.command
        self.subrpocess_call(cmd)

    def subrpocess_call(self, params, shell=True):
        """ Run command or command list with arguments.  Wait for commands to
            complete
            If args.verbose is true, prints command
            If any errors stop list execution and returns error
            if shell=True go shell mode (only for --cron-jobs)

        :param params: command or command list
        :return: error return
        """
        # if not a list convert to a one element list
        params = params if isinstance(params, list) else [params]

        # traverse list executing shell commands
        for _cmd in params:
            # if shell = True we do no split
            cmd = _cmd if shell else _cmd.split()
            if self._parent.verbose:
                msg.run(' ')
                if shell:
                    msg.run(cmd)
                else:
                    msg.run(' '.join(cmd))
                msg.run(' ')
            ret = subprocess.call(cmd, shell=shell)
            if ret:
                return msg.err('The command %s returned with %s' % (cmd, str(ret)))
    @property
    def args(self):
        return self._args

    @property
    def usr_msg(self):
        return self._usr_msg

    @property
    def command(self):
        return self._command


class CreateGitignore(Command):
    def execute(self):
        # crear el gitignore en el archivo que viene del comando
        values = ['.idea/\n', '*.pyc\n', '__pycache__\n']
        with open(self._command, 'w') as _f:
            for value in values:
                _f.write(value)

    @staticmethod
    def check_args():
        return True

class MakedirCommand(Command):
    def check_args(self):
        # si el directorio existe no lo creamos
        return not os.path.isdir(self._args)

class ExtractSourcesCommand(Command):
    @staticmethod
    def check_args():
        return True

class CloneRepo(Command):
    def check_args(self):
        # si el directorio no existe dejamos clonar
        return not os.path.isdir(self._args)

class PullRepo(Command):
    def check_args(self):
        # si el directorio existe dejamos pulear
        return os.path.isdir(self._args)

class PullImage(Command):
    @staticmethod
    def check_args():
        return True

class CreateNginxTemplate(Command):
    def check_args(self):
        # si el archivo existe no lo dejamos pasar
        return not os.path.isfile(self._args)

    def execute(self):
        # leer el nginx.conf
        with open('/usr/local/nginx.conf', 'r') as _f:
            conf = _f.read()

        # poner el nombre del cliente en el config
        conf = conf.replace('$client$', self._client_name)

        with open(self._command, 'w') as _f:
            _f.write(conf)


class WriteConfigFile(Command):
    def check_args(self):
        return True

    def check_item(self, search_item, search_list):
        for item in search_list:
            if search_item in item:
                return item
        return False

    def execute(self):
        # obtener el cliente a partir del nombre
        arg = self._args
        client = arg['client']

        # obtener los repositorios que hay en sources, para eso se recorre souces y se
        # obtienen todos los directorios que tienen un .git adentro.
        repos = list()
        sources = client.sources_dir
        for root, dirs, files in os.walk(sources):
            # si tiene un directorio .git es un repositorio
            if '.git' in dirs:
                repos.append(root.replace(sources,''))
            # si tiene un archivo .git es un subdirectorio
            if '.git' in files:
                repos.append(root.replace(sources, ''))

        repos = ['/opt/odoo/custom-addons/' + x for x in repos]
        repos = ','.join(repos)

        # obtener la configuracion definida en el manifiesto
        conf = client.config or []

        # Actualizar el archivo odoo.conf

        # Leer el archivo de configuracion original
        odoo_conf = OdooConf(client.config_file)
        odoo_conf.read_config()
        odoo_conf.add_list_data(conf)

        # siempre sobreescribimos estas tres cosas.
        odoo_conf.add_line('addons_path = %s' % repos)
        odoo_conf.add_line('unaccent = True')
        odoo_conf.add_line('data_dir = /opt/odoo/data')

        # si estoy en modo debug, sobreescribo esto
        if client.debug:
            odoo_conf.add_line('workers = 0')
            odoo_conf.add_line('max_cron_threads = 0')
            odoo_conf.add_line('limit_time_cpu = 0')
            odoo_conf.add_line('limit_time_real = 0')
        else:
            # no estoy en modo debug,
            # si no defino workers en el manifiesto lo calculo
            line = self.check_item('workers', conf)
            if not line:
                # Calculo los workers
                # You should use 2 worker threads CPU
                odoo_conf.add_line('workers = %s' % (os.cpu_count() * 2))
            else:
                odoo_conf.add_line(line)

            # si no defino cron_threads en el manifiesto lo calculo
            line = self.check_item('max_cron_threads', conf)
            if not line:
                # Calculo los cron threads
                # You should use 1 cron thread per available CPU
                odoo_conf.add_line('max_cron_threads = %s' % os.cpu_count())
            else:
                odoo_conf.add_line(line)

        odoo_conf.write_config()

        # Corregir los permisos de odoo.conf
        os.chmod(client.config_file, stat.S_IREAD + stat.S_IWRITE + stat.S_IWOTH + stat.S_IROTH)


class MessageOnly(Command):
    @staticmethod
    def check_args():
        """ Siempre lo dejamos pasar
        """
        return True

    @staticmethod
    def execute():
        pass



# -*- coding: utf-8 -*-


class Repo(object):
    def __init__(self, value):
        self._dict = value

    @property
    def name(self):
        return self._dict.get('repo')

    @property
    def dir_name(self):
        return self._dict.get('repo')

    @property
    def branch(self):
        return self._dict.get('branch')

    @property
    def url(self):
        if self._dict.get('ssh', False):
            template = 'git@{}:{}/{}'
        else:
            template = 'https://{}/{}/{}'

        return template.format(self._dict.get('host', 'github.com'),
                               self._dict.get('usr'),
                               self._dict.get('repo'))

    @property
    def formatted(self):
        aaa = self._dict['usr'] + '/' + self._dict['repo']
        ret = 'b ' + self._dict['branch'].ljust(7) + ' ' + aaa.ljust(30)
        return ret

    @property
    def clone(self):
        return 'clone --depth 1 -b {} {}'.format(self.branch, self.url)

    @property
    def pull(self):
        return 'pull'


class Repo2(object):
    def __init__(self, value, branch):
        """ Sintaxis <repo> [<directory>[/<directory>] [-b <branch>]
            El branch debe estar despues del repo, si no esta se toma el branch
            que viene como parametro, si no viene nada es una excepcion.
            El directorio va despues del repo y puede no estar
        """
        # parsear value en una lista
        parsed = value.split(' ')
        # eliminar los espacios
        parsed = [i for i in parsed if i != '']
        # obtener el branch si es que existe
        if '-b' in parsed:
            index = parsed.index('-b')
            self._branch = parsed[index + 1]
            # eliminar el -b y el parametro branch
            parsed.remove('-b')
            parsed.remove(self._branch)
        else:
            self._branch = branch

        self._url = parsed[0]

        # si me quedan dos parametros tengo un directorio
        if len(parsed) > 1:
            self._dir = parsed[1]
            self._extra_dir = True
        else:
            parsed = self._url.split('/')
            self._dir = parsed[len(parsed) -1].replace('.git', '')
            self._extra_dir = False

    @property
    def dir_name(self):
        """ Obtener el directorio donde se pone el repo
        """
        return self._dir

    @property
    def branch(self):
        return self._branch

    @property
    def url(self):
        if self._extra_dir:
            return '%s %s' % (self._url, self._dir)
        else:
            return self._url

    @property
    def formatted(self):
        if self._extra_dir:
            return 'b %s     %s >> %s' % (self.branch, self._url, self._dir)
        else:
            return 'b %s     %s' % (self.branch, self.url)


    @property
    def clone(self):
        return 'clone --depth 1 -b %s %s' % (self.branch, self.url)

    @property
    def pull(self):
        return 'pull'



import os
import ast
from odoo_env.messages import Msg
from odoo_env.constants import BASE_DIR, SERVER_BASE_DIR
from odoo_env.repos import Repo, Repo2
from odoo_env.images import Image, Image2
from odoo_env.config import OeConfig

msg = Msg()


class Client(object):
    def __init__(self, odooenv, name):
        """ Busca el cliente en la estructura de directorios, pero si no lo
            encuentra pide un directorio donde esta el repo que lo contiene
        """
        # parent es siempre un objeto OdooEnv
        self._parent = odooenv
        self._name = name
        self._license = False
        self._images = []
        self._repos = []
        self._port = False
        self._version = ''

        # si estamos en test accedo a data
        if name[0:5] in ['test_', 'test2']:
            path = os.path.dirname(os.path.abspath(__file__))
            path = path.replace('odoo_env', 'odoo_env/data')
            manifest = self.get_manifest(path)
            OeConfig().save_client_path(name, path)
        else:
            manifest = self.get_manifest(BASE_DIR)
        if not manifest:
            msg.inf('Can not find client %s in this host installation.\n'
                    'We will try in current dir' % self._name)

            # mantener compatibilidad con python2
            import six
            six.moves.input('Hit Enter to continue or CTRL C to exit')
            manifest, _ = self.get_manifest_from_struct(os.getcwd())
            if not manifest:
                msg.err('Can not find client %s in current dir' % name)

            msg.inf('Client found!')
            msg.inf('Name %s\nversion %s\n' % (manifest.get('name'),
                                               manifest.get('version')))

        self.check_common(manifest)

        # verificar version del manifiesto
        ver = manifest.get('env-ver', '1')
        if ver == '1':
            self.check_v1(manifest)
            msg.warn('The manifest syntax is deprecated, please upgrade to env-ver 2')
            msg.warn('see documentation at https://jobiols.github.io/odoo-env/')
            msg.warn(' ')
        elif ver == '2':
            self.check_v2(manifest)
        else:
            msg.err('Not supported syntax version in manifest, please set env-ver to '
                    '1 or 2')

    def check_v1(self, manifest):
        # Chequar que el manifiesto tenga bien las cosas
        if not manifest.get('docker'):
            msg.err('No images in manifest %s' % self.name)

        if not manifest.get('repos'):
            msg.err('No repos in manifest %s' % self.name)

        # Crear imagenes y repos
        self._repos = []
        for rep in manifest.get('repos'):
            self._repos.append(Repo(rep))

        self._images = []
        for img in manifest.get('docker'):
            self._images.append(Image(img))

    def check_v2(self, manifest):
        # Chequar que el manifiesto tenga bien las cosas
        if not manifest.get('docker-images'):
            msg.err('No images in manifest %s please '
                    'add a docker-images key' % self.name)

        if not manifest.get('git-repos'):
            msg.err('No repos in manifest %s please add a git-repos key' % self.name)

        # leer si es enterprise o community, default community
        self._license = manifest.get('odoo-license', 'CE')

        if self._license not in ['EE', 'CE']:
            msg.err('License must be EE or CE')

        # Crear imagenes y repos
        for rep in manifest.get('git-repos'):
            self._repos.append(Repo2(rep, self._version))

        for img in manifest.get('docker-images'):
            self._images.append(Image2(img))

        # levantar el nombre del user server
        self._prod_server = manifest.get('prod_server', 'ubuntu')

    def check_common(self, manifest):
        self._port = manifest.get('port', 8069)
        self._longpolling_port = manifest.get('longpolling_port', 8072)
        self._external_dependencies = manifest.get('external_dependencies', {})
        ver = manifest.get('version')
        if not ver:
            msg.err('No version tag in manifest %s' % self.name)

        _x = ver.find('.') + 1
        _y = ver[_x:].find('.') + _x
        self._version = ver[0:_y]

        name = manifest.get('name').lower()
        if not self._name == name.split()[0]:
            msg.err('You intend to install client %s but in manifest, '
                    'the name is %s' % (self._name, manifest.get('name')))

        # Tomar los datos para odoo.conf
        self.config = manifest.get('config', [])

    def get_manifest_from_struct(self, path):
        """ leer un manifest que esta dentro de una estructura de directorios
            revisar toda la estructura hasta encontrar un manifest.
            devolver el manifest y el path
        """
        for root, dirs, files in os.walk(path):
            set_files = set(['__openerp__.py', '__manifest__.py']).intersection(files)
            for file in list(set_files):
                manifest_file = '%s/%s' % (root, file)
                manifest = self.load_manifest(manifest_file)
                name = manifest.get('name', False)
                if name and name.lower() == self._name:
                    return manifest, root
        return False, False

    def get_manifest(self, path):
        """
        :param path: path base para buscar el cliente
        :return: manifiesto del cliente
        """
        # traer el path al cliente de la configuracion
        client_path = OeConfig().get_client_path(self._name)
        # si lo encuentro traigo el manifest rapidamente con el path
        if client_path:
            manifest, _ = self.get_manifest_from_struct(client_path)
            return manifest
        else:
            # no lo encuentro, busco en toda la estructura de directorios
            manifest, path = self.get_manifest_from_struct(path)
            if manifest:
                # si lo encuentro lo guardo en el archivo para la proxima
                OeConfig().save_client_path(self._name, path)
            # devuelvo el manifiesto o false si no esta
            return manifest

    @staticmethod
    def load_manifest(filename):
        """
        Loads a manifest
        :param filename: absolute filename to manifest
        :return: manifest in dictionary format
        """
        manifest = ''
        with open(filename, 'r') as _f:
            for line in _f:
                if line.strip() and line.strip()[0] != '#':
                    manifest += line
            try:
                ret = ast.literal_eval(manifest)
            except Exception:
                return {'name': 'none'}
            return ret

    def image(self, image_name):
        for img_dict in self._images:
            if img_dict.get('name') == image_name:
                img = img_dict.get('img')
                ver = img_dict.get('ver')
                ret = img_dict.get('usr')
                if img:
                    ret += '/' + img
                if ver:
                    ret += ':' + ver
                return ret
        msg.err('There is no %s image found in this manifest' % image_name)

    def get_image(self, value):
        for image in self._images:
            if image.short_name == value:
                return image
        return False

    @property
    def name(self):
        return self._name

    @property
    def version(self):
        return self._version

    @property
    def numeric_ver(self):
        return float(self.version[0:2])

    @property
    def repos(self):
        return self._repos

    @property
    def images(self):
        return self._images

    @property
    def port(self):
        return self._port

    @property
    def external_dependencies(self):
        return self._external_dependencies

    @property
    def longpolling_port(self):
        return self._longpolling_port

    @property
    def version_dir(self):
        """ /odoo_ar/odoo-13.0/
            /odoo_ar/odoo-13.0e/
        """
        lic = 'e' if self._license == 'EE' else ''
        return '%sodoo-%s%s/' % (BASE_DIR, self._version, lic)

    @property
    def server_version_dir(self):
        """ /odoo_ar/odoo-13.0/
            /odoo_ar/odoo-13.0e/
        """
        lic = 'e' if self._license == 'EE' else ''
        return '%sodoo-%s%s/' % (SERVER_BASE_DIR, self._version, lic)

    @property
    def base_dir(self):
        """ /odoo_ar/odoo-13.0/clientname/
            /odoo_ar/odoo-13.0e/clientname/
        """
        return '%s%s/' % (self.version_dir, self._name)

    @property
    def server_base_dir(self):
        """ /odoo_ar/odoo-13.0/clientname/
            /odoo_ar/odoo-13.0e/clientname/
        """
        return '%s%s/' % (self.server_version_dir, self._name)

    @property
    def backup_dir(self):
        """ /odoo_ar/odoo-13.0/clientname/backup_dir/
        """
        return self.base_dir + 'backup_dir/'

    @property
    def server_backup_dir(self):
        """ /odoo_ar/odoo-13.0/clientname/backup_dir/
        """
        return self.server_base_dir + 'backup_dir/'

    @property
    def sources_dir(self):
        """ /odoo_ar/odoo-13.0/clientname/sources/
        """
        return self.base_dir + 'sources/'

    @property
    def psql_dir(self):
        """ /odoo_ar/odoo-13.0/clientname/postgresql/
        """
        return self.base_dir + 'postgresql/'

    @property
    def config_file(self):
        """ /odoo_ar/odoo-13.0/clientname/config/odoo.conf
        """
        return self.base_dir + 'config/odoo.conf'

    @property
    def nginx_dir(self):
        """ /odoo_ar/nginx/
        """
        return '%snginx/' % BASE_DIR

    @property
    def debug(self):
        return self._parent.debug

    @property
    def prod_server(self):
        return self._prod_server




import os
from datetime import datetime
import json
import yaml
import tornado
import tornado.httpclient
import tornado.options
import tornado.process
import tornado.web
import tornado.websocket
from odoo_env.__init__ import __version__
from odoo_env.messages import Msg

USER_CONFIG_PATH = os.path.expanduser('~') + '/.config/oe/'
USER_CONFIG_FILE = USER_CONFIG_PATH + 'oe_config.yaml'
USER_CONFIG_FILE_TEST = USER_CONFIG_PATH + 'oe_config_test.yaml'

oe_config = False

_instances = {}


class Singleton(object):
    def __new__(cls, *args, **kw):
        if cls not in _instances:
            instance = super().__new__(cls)
            _instances[cls] = instance
        return _instances[cls]


class OeConfig(Singleton):

    @staticmethod
    def get_config_data():
        template = {'clients': []}
        # obtener el archivo con los datos de clientes
        try:
            with open(USER_CONFIG_FILE, 'r') as config:
                ret = yaml.safe_load(config)
        except Exception:
            return template
        return ret if ret else template

    def save_config_data(self, config):
        """ Salvar el conjunto de paths a los clientes
        """""
        # chequear si esta el archivo y sino crear el path
        if not os.path.exists(USER_CONFIG_PATH):
            os.makedirs(USER_CONFIG_PATH)

        with open(USER_CONFIG_FILE, 'w') as config_file:
            yaml.dump(config, config_file, default_flow_style=False,
                      allow_unicode=True)

    def get_base_dir(self):
        config = self.get_config_data()
        return config.get('base_dir', '/odoo_ar/')

    def get_client_path(self, client_name):
        """ Traer el path de un cliente
        """
        config = self.get_config_data()

        clients = config.get('clients', False)

        for client in clients:
            if client.get(client_name):
                return client.get(client_name)
        return False

    def save_client_path(self, client_name, path):
        """ Salvar el path al cliente, una sola vez
        """
        if not self.get_client_path(client_name):
            # me traigo la configuracion
            config = self.get_config_data()
            # obtengo lista de clientes
            client_list = config['clients']
            # agrego el cliente
            client_list.append({client_name: path})
            # salvo la configuracion
            self.save_config_data(config)

    def get_client(self):
        config = self.get_config_data()
        return config.get('client', False)

    def save_client(self, client):
        config = self.get_config_data()
        config['client'] = client
        self.save_config_data(config)

    def get_environment(self):
        """ Traer el ambiente con prod por defecto
        """
        config = self.get_config_data()
        return config.get('environment', 'prod')

    def save_environment(self, environment):
        """ Salvar el ambiente
        """
        config = self.get_config_data()
        config['environment'] = environment
        self.save_config_data(config)

    def check_version(self):
        # me traigo la configuracion

        config = self.get_config_data()
        dt_today = datetime.today()

        # veo las fechas, si no tiene fecha es que esta recien instalado
        # me guardo la fecha y termino
        last_check = config.get('last_version_check', False)
        if not last_check:
            config['last_version_check'] = dt_today.strftime('%Y-%m-%d')
            self.save_config_data(config)
            return True

        # tiene fecha, la paso a datetime
        dt_last = datetime.strptime(last_check, '%Y-%m-%d')

        # verifico la version cada 10 dias
        if abs((dt_today - dt_last).days) > 10:
            # guardo la fecha del chequeo
            config['last_version_check'] = dt_today.strftime('%Y-%m-%d')
            self.save_config_data(config)

            http = tornado.httpclient.HTTPClient()
            try:
                response = http.fetch(
                    'https://pypi.python.org/pypi/odoo-env/json',
                    connect_timeout=1,
                    request_timeout=1,
                )
                info = json.loads(response.buffer.read().decode('utf-8'))
                version = info['info']['version']
                if version != __version__:
                    Msg().warn('WARNING you are using odoo-env version %s '
                               'however version %s is '
                               'available ' % (__version__, version))
                    Msg().warn('You should consider upgrading via the "pip3 '
                               'install --upgrade odoo-env" command.')
            except Exception as ex:
                Msg().inf('Can not check odoo-env Version, do you have internet '
                          'connectivity?')

        return True



import configparser
from odoo_env import __version__

class OdooConf(object):
    """ Representa el archivo odoo.conf y sus metodos para leer
        salvar y agregarle datos
    """

    def __init__(self, filename):
        self._filename = filename
        self.config = dict()

    def read_config(self):
        """ lee el odoo.conf del archivo y lo vuelca a un diccionario
        """
        config = configparser.ConfigParser()
        config.read(self._filename)
        self.config = config._sections

    def write_config(self):
        """ Escribe o sobreescribe el odoo.conf file
        """
        config = configparser.ConfigParser(allow_no_value=True)
        config.add_section('options')
        config.set('options', '; Configuration file generated by odoo-env v%s' % __version__)
        config.read_dict(self.config)
        with open(self._filename, 'w') as _fp:
            config.write(_fp)

    @staticmethod
    def _parse(line):
        """ dado un string de la forma 'key = value'
            devuelve key y value
        """
        splited = line.split('=')
        key = splited[0].strip()
        value = splited[1].strip()
        return key, value

    def add_list_data(self, datas):
        """ Agrega al config parametros de una lista con la sintaxis que tendra
            finalmente el archivo. o sea 'key = value'
        """
        for data in datas:
            self.add_line(data)

    def add_line(self, line):
        """ Agrega al config una linea de configuración, con la forma 'key = value'
        """
        key, value = self._parse(line)
        if 'options' not in self.config:
            self.config['options'] = dict()
        self.config['options'][key] = value
